%-*-mode:prolog-*-

%%%
%%% metacircular.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for a metacircular Prolog interpreter.
%%%

%%%
%%% demo(+Goal, +Clauses)
%%%
%%% Succeeds if we can prove that Goal is a logical consequence of Clauses.
%%%
%%% This is an interpreter for pure Prolog. It uses the following built-in
%%% predicates:
%%%   - copy_term/2
%%%   - true/0
%%%
%%% Since the intention is for demo/2 to be applied to itself, we implement
%%% these built-in predicates here. 
%%%
%%% The reason for copy_term/2 is because we need a fresh copy of the clauses
%%% when we search to find a clause head that matches Goal. 
%%% 
%%% Although demo/2 uses member/2, the predicate member/2 doesn't have to be a 
%%% built-in predicate, we can include a definition for it in Clauses. 
%%% This is what we will do below.
%%%
demo((Goal1, Goal2), Clauses) :-             %%% Conjunction.
	demo(Goal1, Clauses),
	demo(Goal2, Clauses).
demo((Goal1; Goal2), Clauses) :-             %%% Disjunction.
	demo(Goal1, Clauses)
    ;
	demo(Goal2, Clauses).
demo(true, _) :- true.                       %%% Built-in predicate true/0.
demo(copy_term(A, B), _) :-                  %%% Built-in predicate copy_term/2
	copy_term(A, B).
demo(Goal, Clauses) :-                       %%% Look up a clause
	copy_term(Clauses, ClausesCopy),
	member((Goal :- Body), ClausesCopy),
	demo(Body, Clauses).

%%%
%%% metacircular_demo(+Goal)
%%%
%%% Succeeds if we can prove that we can prove Goal.
%%%
%%% This is the interpreter top-level. We collect all of the clauses for demo/2 
%%% and add in clauses for member/2. We then pass these these along with Goal to 
%%% demo/2 via demo/2, i.e we use demo/2 to interpret demo/2.
%%%
metacircular_demo(Goal) :-
	findall((demo(G, C) :- Body), clause(demo(G, C), Body), Clauses),
	AllClauses = [(member(A, [A|_]) :- true), (member(A, [_|T]) :- member(A, T))|Clauses],
	demo(demo(Goal, AllClauses), AllClauses).
