%-*-mode:prolog-*-

%%%
%%% pl_clausal_form.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains the procedures to turn a formula in Conjunctive
%%% Normal Form (CNF) to one in clausal form.
%%%
:- ensure_loaded(examples(pl_operators)).
:- ensure_loaded(runtime(ordset)).

%%%
%%% pl_simplify1(F, S).
%%%
%%% S is a rewriting of F to make it simpler.
%%%
pl_simplify1(~true, false) :- !.
pl_simplify1(~false, true) :- !.
pl_simplify1(~ ~Phi, Phi) :- !.
pl_simplify1(_ /\ false, false) :- !.
pl_simplify1(false /\ _, false) :- !.
pl_simplify1(Phi /\ true, Phi) :- !.
pl_simplify1(true /\ Phi, Phi) :- !.
pl_simplify1(Phi \/ false, Phi) :- !.
pl_simplify1(false \/ Phi, Phi) :- !.
pl_simplify1(_ \/ true, true) :- !.
pl_simplify1(true \/ _, true) :- !.
pl_simplify1(Phi, Phi).

%%%
%%% pl_simplify(F, S)
%%%
%%% S is the simplification of the formula F which we assume is in
%%% CNF
%%%
pl_simplify(~Phi, S) :- 
	!,
	pl_simplify(Phi, PhiSimplified), 
	pl_simplify1(~PhiSimplified, S). 
pl_simplify(Phi /\ Psi, S) :-
	!,
	pl_simplify(Phi, PhiSimplified),
	pl_simplify(Psi, PsiSimplified),
	pl_simplify1(PhiSimplified /\ PsiSimplified, S).
pl_simplify(Phi \/ Psi, S) :-
	!,
	pl_simplify(Phi, PhiSimplified),
	pl_simplify(Psi, PsiSimplified),
	pl_simplify1(PhiSimplified \/ PsiSimplified, S).
pl_simplify(Phi, Phi).

%%%
%%% pl_clausal_form(+A, -B)
%%%
%%% B is the clausal form of the formula A in CNF.
%%%
pl_clausal_form(A, Result) :-
	pl_simplify(A, SimplifiedA),
	pl_clausal_form_conjunction(SimplifiedA, Result).

%%%
%%% pl_clausal_form_conjunction(A, B)
%%%
%%% Process the conjuncts of A giving the set of sets B.
%%%
pl_clausal_form_conjunction(A /\ B, Result) :-
	!,
	pl_clausal_form_conjunction(A, ResultA),
	pl_clausal_form_conjunction(B, ResultB),
	ord_union(ResultA, ResultB, Result).
pl_clausal_form_conjunction(true, []) :- !.
pl_clausal_form_conjunction(A, [Result]) :-
	pl_clausal_form_disjunction(A, Result).

%%%
%%% pl_clausal_form_disjunction(A, B)
%%%
%%% Process the disjuncts of A giving the set B.
%%%
pl_clausal_form_disjunction(A \/ B, Result)  :-
	!,
	pl_clausal_form_disjunction(A, ResultA),
	pl_clausal_form_disjunction(B, ResultB),
	ord_union(ResultA, ResultB, Result).
pl_clausal_form_disjunction(false, []) :- !.
pl_clausal_form_disjunction(A, [A]).
