%-*-mode:prolog-*-

%%%
%%% pl_cnf.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the transformation of formulas in propositional logic
%%% to Conjunctive Normal Form (CNF).
%%%
:- ensure_loaded(examples(pl_operators)).

%%%
%%% cnf_rewrite_connectives(+A, ?B)
%%%
%%% The formula A is rewritten to B so that an implication
%%%   Phi => Psi
%%% is replaced by
%%%   ~Phi \/ Psi
%%% and an equivalance
%%%   Phi <=> Psi
%%% is replaced by
%%%   (~Phi \/ Psi) /\ (Phi \/ ~Psi)
%%%
cnf_rewrite_connectives(~A0, ~A1) :- 
	!,
	cnf_rewrite_connectives(A0, A1).
cnf_rewrite_connectives(A0 /\ B0, A1 /\ B1) :-
	!,
	cnf_rewrite_connectives(A0, A1),
	cnf_rewrite_connectives(B0, B1).
cnf_rewrite_connectives(A0 \/ B0, A1 \/ B1) :-
	!,
	cnf_rewrite_connectives(A0, A1),
	cnf_rewrite_connectives(B0, B1).
cnf_rewrite_connectives(A0 => B0, ~A1 \/ B1) :-
	!,
	cnf_rewrite_connectives(A0, A1),
	cnf_rewrite_connectives(B0, B1).
cnf_rewrite_connectives(A0 <=> B0, (~A1 \/ B1) /\ (A1 \/ ~B1)) :-
	!,
	cnf_rewrite_connectives(A0, A1),
	cnf_rewrite_connectives(B0, B1).
cnf_rewrite_connectives(A, A).

%%%
%%% cnf_push_negation(+A, -B)
%%%
%%% The formula A is rewritten to B using the following equivalences
%%%   ~(~A) <=> A
%%%   ~(A /\ B) <=> ~A \/ ~B  (De Morgan)
%%%   ~(A \/ B) <=> ~A /\ ~B  (De Morgan)
%%%   ~false <=> true
%%%   ~true <=> false
%%%
cnf_push_negation(~(~A0), A1) :- 
	!,
	cnf_push_negation(A0, A1).
cnf_push_negation(~(A0 /\ B0), A1 \/ B1) :-
	!,
	cnf_push_negation(~A0, A1),
	cnf_push_negation(~B0, B1).
cnf_push_negation(~(A0 \/ B0), A1 /\ B1) :-
	!,
	cnf_push_negation(~A0, A1),
	cnf_push_negation(~B0, B1).
cnf_push_negation(A0 /\ B0, A1 /\ B1) :-
	!,
	cnf_push_negation(A0, A1),
	cnf_push_negation(B0, B1).
cnf_push_negation(A0 \/ B0, A1 \/ B1) :-
	!,
	cnf_push_negation(A0, A1),
	cnf_push_negation(B0, B1).
cnf_push_negation(~ true, false).
cnf_push_negation(~ false, true).
cnf_push_negation(A, A).

%%%
%%% cnf_distribute(+A, ?B)
%%%
%%% The formula A is rewritten to B by taking a subterm of the form
%%%    A \/ (B /\ C), or (A /\ B) \/ C
%%% and distributed the disjunction over the conjunction.
%%%
cnf_distribute(A \/ (B /\ C), (A \/ B) /\ (A \/ C)) :- !.
cnf_distribute((A /\ B) \/ C, (A \/ C) /\ (B \/ C)) :- !.
cnf_distribute(A0 /\ B, (A1 /\ B)) :-
	cnf_distribute(A0, A1).
cnf_distribute(A /\ B0, (A /\ B1)) :-
	cnf_distribute(B0, B1).
cnf_distribute(A0 \/ B, (A1 \/ B)) :-
	cnf_distribute(A0, A1).
cnf_distribute(A \/ B0, (A \/ B1)) :-
	cnf_distribute(B0, B1).

%%%
%%% cnf_distribute_loop(+A, +B)
%%%
%%% The formula A is successively rewritten to B using
%%% cnf_distribute/2.
%%%
cnf_distribute_loop(A, C) :-
	cnf_distribute(A, B),
	!,
	cnf_distribute_loop(B, C).
cnf_distribute_loop(A, A).

%%%
%%% cnf_transform(+A, ?B)
%%%
%%% Formula B is a conjunctive normal form of formula A.
%%%
cnf_transform(A, D) :-
	cnf_rewrite_connectives(A, B),
	cnf_push_negation(B, C),
	cnf_distribute_loop(C, D).
	


