%-*-mode:prolog-*-

%%%
%%% pl_definitional_cnf.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the transformation of formulas in propositional logic
%%% to definitional Conjunctive Normal Form (CNF).
%%%
%%% The method shown here is from the "Handbook of Practical Logic and Automated Reasoning" 
%%% by John Harrison.
%%%

:- ensure_loaded(examples(pl_operators)).
:- ensure_loaded(examples(pl_nenf)).
:- ensure_loaded(examples(pl_cnf)).

%%%
%%% Definitional CNF
%%%
%%% The transformation to CNF can lead to an explosion in formula size. Definitional CNF
%%% is an attempt to have a CNF without this big increase in size in the worst case.
%%%
%%% The idea is to transform a formula into an equisatisfiable one which should be smaller than the 
%%% equivalent CNF. Note: equisatisfiable NOT equivalent.
%%% If x does not occur in q then p{x/q} and (x <=> q)/\p are equisatisfiable.
%%% It is called "definitional" as subterms are used to define new variables which replace them.
%%%

%%%
%%% Here we have three types of definitional CNF
%%%
%%% 1) Ordinary definitional CNF.
%%% 2) Optimised definitional CNF.
%%% 3) Optimised definitional 3-CNF.
%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                                %%%
%%%   Ordinary Definitional CNF    %%%
%%%                                %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%
%%% pl_make_prop_variable(+Id, +Number, ?NewVariable, ?NewNumber)
%%%
%%% Succeeds if NewVariable has the name which is the concatenation of the names of
%%% Id and Number, in that order, and NewNumber is Number+1.
%%%
pl_make_prop_variable(Id, Number, NewVariable, NewNumber) :-
	name(Number, NumberCodes),
	name(Id, IdCodes),
	append(IdCodes, [0'_|NumberCodes], NewVariableCodes),
	name(NewVariable, NewVariableCodes),
	NewNumber is Number+1.


%%%
%%% pl_max_varindex(+PrefixAtom, +N, +Var, ?Max)
%%%
%%% Succeeds if Max is the maximum of N and M where M is the number represented by the codes
%%% found in the name of Var after the codes of PrefixAtom.
%%%
pl_max_varindex(PrefixAtom, N, Var, Max) :-
	name(PrefixAtom, PrefixAtomName),
	name(Var, VarName),
	(   append(PrefixAtomName, VarPostfix, VarName), name(VarNumber, VarPostfix), number(VarNumber) ->
	    max(N, VarNumber, Max)
	;   Max = N
        ).

%%%
%%% pl_formula_max_varindex(+F, +P, +I, ?O)
%%%
%%% Succeeds if the maximum index of variables with prefix P in formula F is
%%% O. Here I is the initial index, usually 0.
%%%
pl_formula_max_varindex(Atom, Prefix, In, Out) :- 
	atom(Atom), 
	!,
	pl_max_varindex(Prefix, In, Atom, Out).
pl_formula_max_varindex(~Phi, Prefix, In, Out) :-
	pl_formula_max_varindex(Phi, Prefix, In, Out).
pl_formula_max_varindex(Phi /\ Psi, Prefix, In, Out) :-
	pl_formula_max_varindex(Phi, Prefix, In, Mid),
	pl_formula_max_varindex(Psi, Prefix, Mid, Out).
pl_formula_max_varindex(Phi \/ Psi, Prefix, In, Out) :-
	pl_formula_max_varindex(Phi, Prefix, In, Mid),
	pl_formula_max_varindex(Psi, Prefix, Mid, Out).
pl_formula_max_varindex(Phi => Psi, Prefix, In, Out) :-
	pl_formula_max_varindex(Phi, Prefix, In, Mid),
	pl_formula_max_varindex(Psi, Prefix, Mid, Out).
pl_formula_max_varindex(Phi <=> Psi, Prefix, In, Out) :-
	pl_formula_max_varindex(Phi, Prefix, In, Mid),
	pl_formula_max_varindex(Psi, Prefix, Mid, Out).


%%%
%%% pl_definitional_cnf_map_cnf(+F, +Acc, ?Result)
%%% Create a CNF for each member of the list of input forumulae F and conjoin
%%% them altogether to create Result. Acc is the accumulated conjuncts.
%%%
%%% Since CNF transformation is an expensive operation, doing it conjunct 
%%% by conjunct is easier than just conjoining them all and calling cnf_transform/2
%%% once.
%%%
pl_definitional_cnf_map_cnf([], Conjuncts, Conjuncts).
pl_definitional_cnf_map_cnf([Phi|Phis], Acc, Conjuncts) :-
	cnf_transform(Phi, Conjunct),
	pl_definitional_cnf_map_cnf(Phis, Conjunct /\ Acc, Conjuncts).

%%%
%%% pl_definitional_cnf(+Phi, +Din, +Nin, ?Psi, ?Dout, ?Nout)
%%%
%%% This is used to build a definition.
%%%
%%% If Phi is a compound NENF formula, Din a set of variable-definition pairs, and Nin the 
%%% maximum var index in Phi, then Psi is a variable with index N+1, Dout is the updated
%%% set of variable-definition pairs, and Nout is the updated max index.  
%%% Otherwise Ps=Phi, Din=Dout, Nin=Nout.
%%%
pl_definitional_cnf(FormIn, Defns, N, FormOut, Defns4, N4) :-
	compound(FormIn),
	FormIn =.. [Op, Phi, Psi],
	!,
	member(Op, [(\/), (/\), (<=>)]),
	NewForm =.. [Op, Phi2, Psi2],
	pl_definitional_cnf(Phi, Defns, N, Phi2, Defns2, N2),
	pl_definitional_cnf(Psi, Defns2, N2, Psi2, Defns3, N3),
	(   member(Var-NewForm, Defns) ->
	    FormOut = Var,
	    Defns4 = Defns3,
	    N3 = N4
	;
	    pl_make_prop_variable('p', N3, NewVar, N4),
	    FormOut = NewVar,
	    Defns4 = [NewVar-NewForm|Defns3]
	).
pl_definitional_cnf(Phi, Defns, N, Phi, Defns, N).

%%%
%%% pl_definitional_cnf(+Phi, ?Psi)
%%%
%%% Succeeds if Psi is the definitial CNF of Phi.
%%%
%%% Try: pl_definitional_cnf((p \/ (q /\ ~r)) /\ s, D).	
%%%
pl_definitional_cnf(PhiIn, PhiOut) :-
	pl_formula_max_varindex(PhiIn, 'p', 0, N),
	M is N+1,
	nenf_transform(PhiIn, PhiNENF),
	pl_definitional_cnf(PhiNENF, [], M, PhiDef, Defns, _),
	findall((A<=>B), member(A-B, Defns), Lst),
	pl_definitional_cnf_map_cnf(Lst, PhiDef, PhiOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                                %%%
%%%   Optimised Definitional CNF   %%%
%%%                                %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Note that we can avoid unnecessary computations if the top-level connective is a conjunction.
%%% In this case we just treat each conjunct seperately then conjoin the results. 
%%% Also, if any of these conjuncts are disjunctions, we can ignore any atomic disjuncts as they are 
%%% already in CNF and don't need to part of a definition.

%%%
%%% pl_and_cnf(+Form1, +Defns1, +N1, -Form2, -Defns2, -N2)
%%%
%%% Recurse through the iterated conjunctions of Form1 with Defns1 and N1.
%%% The conjuncts are passed onto pl_or_cnf/6.
%%% Defns1 are the definitions seen so far, and N1 is the next index to use
%%% for a definition variable.
%%% The resulting definitions are DefnsOut and the last index used for a 
%%% definition variable is N2-1. 
%%%
pl_and_cnf(Phi /\ Psi, Defns, N, PhiOut /\ PsiOut, DefnsOut, Nout) :-
	pl_and_cnf(Phi, Defns,N, PhiOut, DefnsMid, Nmid),
	pl_and_cnf(Psi, DefnsMid, Nmid, PsiOut, DefnsOut, Nout).
pl_and_cnf(Form, Denfs, N, FormOut, DefnsOut, Nout) :-
	pl_or_cnf(Form, Denfs, N, FormOut, DefnsOut, Nout).

%%%
%%% pl_or_cnf(+Form1, +Defns1, +N1, -Form2, -Defns2, -N2)
%%%
%%% Recurse through the iterated disjunctions of Form1 with Defns1 and N1.
%%% The disjuncts are passed onto pl_definitional_cnf/6.
%%% Defns1 are the definitions seen so far, and N1 is the next index to use
%%% for a definition variable.
%%% The resulting definitions are DefnsOut and the last index used for a 
%%% definition variable is N2-1. 
%%%
pl_or_cnf(Phi \/ Psi, Defns, N, PhiOut \/ PsiOut, DefnsOut, Nout) :-
	pl_or_cnf(Phi, Defns,N, PhiOut, DefnsMid, Nmid),
	pl_or_cnf(Psi, DefnsMid, Nmid, PsiOut, DefnsOut, Nout).
pl_or_cnf(Form, Denfs, N, FormOut, DefnsOut, Nout) :-
	pl_definitional_cnf(Form, Denfs, N, FormOut, DefnsOut, Nout).

%%%
%%% pl_optimised_definitional_cnf(+Phi, -Psi)
%%%
%%% Psi is Phi in definitional CNF.
%%%
%%% Try: pl_optimised_definitional_cnf((p \/ (q /\ ~r)) /\ s, D).	
%%%
pl_optimised_definitional_cnf(PhiIn, PhiOut) :-
	pl_formula_max_varindex(PhiIn, 'p', 0, N),
	M is N+1,
	nenf_transform(PhiIn, PhiNENF),
	pl_and_cnf(PhiNENF, [], M, PhiDef, Defns, _),
	findall((A<=>B), member(A-B, Defns), Lst),
	pl_definitional_cnf_map_cnf(Lst, PhiDef, PhiOut).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                                %%%
%%%  Optimised Definitional 3-CNF  %%%
%%%                                %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%
%%% The nonoptimised definitional CNF transformation resulted in 3-CNF, that is
%%% each conjunct is a disjunct of at most 3 literals. This isn't the case 
%%% with the optimised definitional CNF. A simple solution is to just 
%%% treat each conjunct seperately and pass the conjuncts to the definitional
%%% CNF transformation routine. Basically, this is the optimised definitional
%%% CNF that skips the pl_or_cnf/6 call.
%%%

%%%
%%% pl_and_cnf3(+Form1, +Defns1, +N1, -Form2, -Defns2, -N2)
%%%
%%% Recurse through the iterated conjunctions of Form1 with Defns1 and N1.
%%% The conjuncts are passed onto pl_definitional_cnf/6.
%%% Defns1 are the definitions seen so far, and N1 is the next index to use
%%% for a definition variable.
%%% The resulting definitions are DefnsOut and the last index used for a 
%%% definition variable is N2-1. 
%%%
pl_and_cnf3(Phi /\ Psi, Defns, N, PhiOut /\ PsiOut, DefnsOut, Nout) :-
	pl_and_cnf3(Phi, Defns,N, PhiOut, DefnsMid, Nmid),
	pl_and_cnf3(Psi, DefnsMid, Nmid, PsiOut, DefnsOut, Nout).
pl_and_cnf3(Form, Denfs, N, FormOut, DefnsOut, Nout) :-
	pl_definitional_cnf(Form, Denfs, N, FormOut, DefnsOut, Nout).

%%%
%%% pl_optimised_definitional_cnf3(+Phi, -Psi)
%%%
%%% Psi is Phi in definitional 3-CNF.
%%%
%%% Try: pl_optimised_definitional_cnf3((p \/ (q /\ ~r)) /\ s, D).
%%%
pl_optimised_definitional_cnf3(PhiIn, PhiOut) :-
	pl_formula_max_varindex(PhiIn, 'p', 0, N),
	M is N+1,
	nenf_transform(PhiIn, PhiNENF),
	pl_and_cnf3(PhiNENF, [], M, PhiDef, Defns, _),
	findall((A<=>B), member(A-B, Defns), Lst),
	pl_definitional_cnf_map_cnf(Lst, PhiDef, PhiOut).
