%-*-mode:prolog-*-

%%%
%%% pl_dnf.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the transformation of formulas in propositional logic
%%% to Disjunctive Normal Form (DNF).
%%%
:- ensure_loaded(examples(pl_operators)).

%%%
%%% dnf_rewrite_connectives(+A, ?B)
%%%
%%% The formula A is rewritten to B so that an implication
%%%   Phi => Psi
%%% is replaced by
%%%   ~Phi \/ Psi
%%% and an equivalance
%%%   Phi <=> Psi
%%% is replaced by
%%%   (~Phi \/ Psi) /\ (Phi \/ ~Psi)
%%%
dnf_rewrite_connectives(~A0, ~A1) :- 
	!,
	dnf_rewrite_connectives(A0, A1).
dnf_rewrite_connectives(A0 /\ B0, A1 /\ B1) :-
	!,
	dnf_rewrite_connectives(A0, A1),
	dnf_rewrite_connectives(B0, B1).
dnf_rewrite_connectives(A0 \/ B0, A1 \/ B1) :-
	!,
	dnf_rewrite_connectives(A0, A1),
	dnf_rewrite_connectives(B0, B1).
dnf_rewrite_connectives(A0 => B0, ~A1 \/ B1) :-
	!,
	dnf_rewrite_connectives(A0, A1),
	dnf_rewrite_connectives(B0, B1).
dnf_rewrite_connectives(A0 <=> B0, (~A1 \/ B1) /\ (A1 \/ ~B1)) :-
	!,
	dnf_rewrite_connectives(A0, A1),
	dnf_rewrite_connectives(B0, B1).
dnf_rewrite_connectives(A, A).

%%%
%%% dnf_push_negation(+A, -B)
%%%
%%% The formula A is rewritten to B using the following equivalences
%%%   ~(~A) <=> A
%%%   ~(A /\ B) <=> ~A \/ ~B  (De Morgan)
%%%   ~(A \/ B) <=> ~A /\ ~B  (De Morgan)
%%%   ~false <=> true
%%%   ~true <=> false
%%%
dnf_push_negation(~(~A0), A1) :- 
	!,
	dnf_push_negation(A0, A1).
dnf_push_negation(~(A0 /\ B0), A1 \/ B1) :-
	!,
	dnf_push_negation(~A0, A1),
	dnf_push_negation(~B0, B1).
dnf_push_negation(~(A0 \/ B0), A1 /\ B1) :-
	!,
	dnf_push_negation(~A0, A1),
	dnf_push_negation(~B0, B1).
dnf_push_negation(A0 /\ B0, A1 /\ B1) :-
	!,
	dnf_push_negation(A0, A1),
	dnf_push_negation(B0, B1).
dnf_push_negation(A0 \/ B0, A1 \/ B1) :-
	!,
	dnf_push_negation(A0, A1),
	dnf_push_negation(B0, B1).
dnf_push_negation(~ true, false).
dnf_push_negation(~ false, true).
dnf_push_negation(A, A).

%%%
%%% dnf_distribute(+A, ?B)
%%%
%%% The formula A is rewritten to B by taking a subterm of the form
%%%    A /\ (B \/ C), or (A \/ B) /\ C
%%% and distributed the conjunction over the disjunction.
%%%
dnf_distribute(A /\ (B \/ C), (A /\ B) \/ (A /\ C)) :- !.
dnf_distribute((A \/ B) /\ C, (A /\ C) \/ (B /\ C)) :- !.
dnf_distribute(A0 /\ B, (A1 /\ B)) :-
	dnf_distribute(A0, A1).
dnf_distribute(A /\ B0, (A /\ B1)) :-
	dnf_distribute(B0, B1).
dnf_distribute(A0 \/ B, (A1 \/ B)) :-
	dnf_distribute(A0, A1).
dnf_distribute(A \/ B0, (A \/ B1)) :-
	dnf_distribute(B0, B1).

%%%
%%% dnf_distribute_loop(+A, +B)
%%%
%%% The formula A is successively rewritten to B using
%%% dnf_distribute/2.
%%%
dnf_distribute_loop(A, C) :-
	dnf_distribute(A, B),
	!,
	dnf_distribute_loop(B, C).
dnf_distribute_loop(A, A).

%%%
%%% dnf_transform(+A, ?B)
%%%
%%% Formula B is a disjunctive normal form of formula A.
%%%
dnf_transform(A, D) :-
	dnf_rewrite_connectives(A, B),
	dnf_push_negation(B, C),
	dnf_distribute_loop(C, D).
	


