%-*-mode:prolog-*-

%%%
%%% pl_dp.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the Davis-Putnam procedure.
%%%
:- ensure_loaded(examples(pl_operators)).
:- ensure_loaded(runtime(ordset)).
:- ensure_loaded(examples(pl_cnf)).
:- ensure_loaded(examples(pl_clausal_form)).

%%%
%%% pl_negate(+A, ?B)
%%%
%%% B is the negation of A.
%%%
pl_negate(~A, A) :- !.
pl_negate(A, ~A).

%%%
%%% pl_dp_one_literal_rule_remove_negated_literals(ClausesIn, P, ClausesOut)
%%%
%%% Succeeds if ClausesOut is ClausesIn with all occurrences of literal P removed.
%%%
%%% Note that each clause in ClausesOut is still in order --- no need to sort.
%%%
pl_dp_one_literal_rule_remove_negated_literals([], _, []).
pl_dp_one_literal_rule_remove_negated_literals([Phi|Phis], Neg, [Psi|Psis]) :- 
	delete_all(Neg, Phi, Psi),
	!,
	pl_dp_one_literal_rule_remove_negated_literals(Phis, Neg, Psis).

%%%
%%% pl_dp_one_literal_rule_remove_clauses(+ClausesIn, +P, +ClausesOut)
%%%
%%% Succeeds if ClausesOut is equal to ClausesIn with all elements that contain
%%% an occurrence of the literal P removed.
%%%
%%% Note that ClausesOut is still in order --- no need to sort.
%%%
pl_dp_one_literal_rule_remove_clauses([], _, []).
pl_dp_one_literal_rule_remove_clauses([Phi|Phis], P, Result) :- 
	member(P, Phi),
	!,
	pl_dp_one_literal_rule_remove_clauses(Phis, P, Result).
pl_dp_one_literal_rule_remove_clauses([Phi|Phis], P, [Phi|Result]) :- 
	pl_dp_one_literal_rule_remove_clauses(Phis, P, Result).

%%%
%%% pl_dp_one_literal_rule(+Phi, -Psi)
%%%
%%% Succeeds if Phi and Psi are equsatisfiable and Phi contains a clause with a single disjunct (literal) and
%%% Psi is Phi where this clause is removed along with the removal of the negation of the literal from all 
%%% other clauses.
%%%
%%% This is also called unit propagation.
%%%
pl_dp_one_literal_rule(Phi, Psi) :-
	member([P], Phi),
	pl_negate(P, NegP),
	pl_dp_one_literal_rule_remove_clauses(Phi, P, Mid),
	pl_dp_one_literal_rule_remove_negated_literals(Mid, NegP,Psi).

%%%
%%% pl_affirmitive_negative_rule_remove_clauses(+ClausesIn, +Pures, -ClausesOut)
%%%
%%% Succeeds if ClausesOut is ClausesIn with all clauses which contain an occurrence of a positive or negative 
%%% literal in Pures removed. Pures only contains positive literals.
%%% 
pl_affirmitive_negative_rule_remove_clauses([], _, []).
pl_affirmitive_negative_rule_remove_clauses([Phi|Phis], Pures, Result) :- 
	member(P, Pures),
	(member(P, Phi) ; member(~P, Phi)),
	!,
	pl_affirmitive_negative_rule_remove_clauses(Phis, Pures, Result).
pl_affirmitive_negative_rule_remove_clauses([Phi|Phis], Pures, [Phi|Result]) :- 
	pl_affirmitive_negative_rule_remove_clauses(Phis, Pures, Result).

%%%
%%% pl_dp_affirmative_negative_rule(Phi, Psi)
%%%
%%% Succeeds if Psi is Phi with all claues that contain a "pure" literal removed.
%%%
%%% We can delete any clause that contains a literal than only occurs either positively or
%%% negatively and not both. Such a literal is called "pure". This preserves satsfiablity.
%%%
%%% Note that for large input clauses you may want to replace the calls to setof with 
%%% somthing that doesn't copy between the clause store and the heap.
%%%
pl_dp_affirmative_negative_rule(Phi, Psi) :-
	setof(Pos, Clause^(member(Clause, Phi), member(Pos, Clause), Pos \= ~_), Positives),
	setof(Neg, Clause^(member(Clause, Phi), member(~Neg, Clause)), Negatives),
	ord_subtract(Positives, Negatives, OnlyPositives),
	ord_subtract(Negatives, Positives, OnlyNegatives),
	append(OnlyPositives, OnlyNegatives, Pures),
	Pures \= [],                                                  % must have them!
	pl_affirmitive_negative_rule_remove_clauses(Phi, Pures, Psi).

%%%
%%% pl_filter_trivial(+ClausesIn,?-ClausesOut)
%%%
%%% ClausesOut is the clausal form ClausesIn with all trivial clauses, that is all clauses
%%% containing both a positive and negative literal with the same atom, removed.
%%%
pl_filter_trivial([], []).
pl_filter_trivial([Clause|Clauses], Result) :-
	member(Lit, Clause),
	pl_negate(Lit, NegLit),
	member(NegLit, Clause),
	!,
	pl_filter_trivial(Clauses, Result).
pl_filter_trivial([Clause|Clauses], [Clause|Result]) :-
	pl_filter_trivial(Clauses, Result).

%%%
%%% combine_all_pairs(+S1, +S2, ?S3)
%%%
%%% Succeeds if S3 is the ordered union of all pairs from the sets S1 and S2
%%%
combine_all_pairs([], S, S) :- !.
combine_all_pairs(S, [], S) :- !.
combine_all_pairs(S1, S2, S4) :- 
	combine_all_pairs(S1, S2, [], S4).

combine_all_pairs([], _, Result, Result).
combine_all_pairs([H|T], S2, Acc, Result) :-
	combine(S2, H, P),
	ord_union(P, Acc, NewAcc),
	combine_all_pairs(T, S2, NewAcc, Result).

combine([], _, []).
combine([H|T], E, [Set|Sets]) :-
	ord_union(E, H, Set),
	combine(T, E, Sets).

%%%
%%% pl_dp_resolution_clauses(+Clauses, +Lit, +NegLit, -PositiveClauses, -NegativeClauses, -OtherClauses)
%%%
%%% Splits Clauses into PositiveClauses which contain Lit and do not contain NegLit, NegativeClauses which
%%% contain NegLit and not Lit, and OtherClauses which contain neither or both.
%%%
%%% Note that all output sets are still ordered.
%%%
pl_dp_resolution_clauses([], _, _, [], [], []).
pl_dp_resolution_clauses([Clause|Clauses], Lit, NegLit, [Clause1|ResultPos], ResultNeg, Others) :-
	member(Lit, Clause),
	\+ member(NegLit, Clause),
	delete_all(Lit, Clause, Clause1),
	!,
	pl_dp_resolution_clauses(Clauses, Lit, NegLit, ResultPos, ResultNeg, Others).	
pl_dp_resolution_clauses([Clause|Clauses], Lit, NegLit, ResultPos, [Clause1|ResultNeg], Others) :-
	member(NegLit, Clause),
	\+ member(Lit, Clause),
	delete_all(NegLit, Clause, Clause1),
	!,
	pl_dp_resolution_clauses(Clauses, Lit, NegLit, ResultPos, ResultNeg, Others).	
pl_dp_resolution_clauses([Clause|Clauses], Lit, NegLit, ResultPos, ResultNeg, [Clause|Others]) :-
	pl_dp_resolution_clauses(Clauses, Lit, NegLit, ResultPos, ResultNeg, Others).	

%%%
%%% pl_dp_resolution_rule(+ClausesIn, ?ClausesOut)
%%%
%%% Succeeds if ClausesOut is ClausesIn subject to a single application of resolution.
%%% Actually, we resolve as much as possible for a certain literal, that is we form all
%%% possible resolvents for the literal.
%%%
pl_dp_resolution_rule(Clauses0, Result) :-
	del(C0, Clauses0, Clauses1),       % Choose a clause C0.
	del(Lit, C0, Alpha),               % Choose a literal from C0. 
	pl_negate(Lit, NegLit),            % Negate this literal.
	\+ member(NegLit, Alpha),          % The negated literal cannot occur in C0.
	del(C1, Clauses1, _),              % Choose another clause C1.
	del(NegLit, C1, Beta),             % The negated literal must occur in C1.
	\+ member(Lit, Beta),              % The original literal cannot occur in C1.
	!,
	pl_dp_resolution_clauses(Clauses0, Lit, NegLit, ClausesPos, ClausesNeg, ClausesOthers),
	combine_all_pairs(ClausesPos, ClausesNeg, Resolvents),   % Form all resolvants for Lit.
	ord_union(Resolvents, ClausesOthers, Mid),
	pl_filter_trivial(Mid, Result).

%%%
%%% pl_dp_loop(+ClausesIn, -ClausesOut)
%%%
%%% Try to apply the following rules in this order to ClausesIn to give
%%% ClausesOut:
%%%   - 1-literal rule (unit propagation).
%%%   - Affirmative Negative rule.
%%%   - Resolution.
%%%
pl_dp_loop(ClausesIn, ClausesOut) :-
	pl_dp_one_literal_rule(ClausesIn, ClausesMid),	
	!,
	pl_dp_loop(ClausesMid, ClausesOut).
pl_dp_loop(ClausesIn, ClausesOut) :-
	pl_dp_affirmative_negative_rule(ClausesIn, ClausesMid),
	!,
	pl_dp_loop(ClausesMid, ClausesOut).
pl_dp_loop(ClausesIn, ClausesOut) :-
	pl_dp_resolution_rule(ClausesIn, ClausesMid),
	!,
	pl_dp_loop(ClausesMid, ClausesOut).
pl_dp_loop(Clauses, Clauses).

%%%
%%% pl_dp_sat(+Phi)
%%%
%%% Succeeds if Phi is satisfiable.
%%%
pl_dp_sat(Phi) :-
	cnf_transform(Phi, CNF),
	pl_clausal_form(CNF, Clauses),
	pl_dp_loop(Clauses, []).

%%%
%%% pl_dp_sat(+Phi)
%%%
%%% Succeeds if Phi is a tautology.
%%%
pl_dp_tautology(Phi) :-
	\+ pl_dp_sat(~Phi).

