%-*-mode:prolog-*-

%%%
%%% pl_dpll.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the Davis-Putnam-Logemann-Loveland procedure.
%%% Compare this file with pl_dp.pl and you will see that the only difference is
%%% that the resolution rule in Davis-Putnam is replaced by the Split rule.
%%%
:- ensure_loaded(examples(pl_operators)).
:- ensure_loaded(runtime(ordset)).
:- ensure_loaded(examples(pl_cnf)).
:- ensure_loaded(examples(pl_clausal_form)).

%%%
%%% pl_negate(+A, ?B)
%%%
%%% B is the negation of A.
%%%
pl_negate(~A, A) :- !.
pl_negate(A, ~A).

%%%
%%% pl_dp_one_literal_rule_remove_negated_literals(ClausesIn, P, ClausesOut)
%%%
%%% Succeeds if ClausesOut is ClausesIn with all occurrences of literal P removed.
%%%
%%% Note that each clause in ClausesOut is still in order --- no need to sort.
%%%
pl_dp_one_literal_rule_remove_negated_literals([], _, []).
pl_dp_one_literal_rule_remove_negated_literals([Phi|Phis], Neg, [Psi|Psis]) :- 
	delete_all(Neg, Phi, Psi),
	!,
	pl_dp_one_literal_rule_remove_negated_literals(Phis, Neg, Psis).

%%%
%%% pl_dp_one_literal_rule_remove_clauses(+ClausesIn, +P, +ClausesOut)
%%%
%%% Succeeds if ClausesOut is equal to ClausesIn with all elements that contain
%%% an occurrence of the literal P removed.
%%%
%%% Note that ClausesOut is still in order --- no need to sort.
%%%
pl_dp_one_literal_rule_remove_clauses([], _, []).
pl_dp_one_literal_rule_remove_clauses([Phi|Phis], P, Result) :- 
	member(P, Phi),
	!,
	pl_dp_one_literal_rule_remove_clauses(Phis, P, Result).
pl_dp_one_literal_rule_remove_clauses([Phi|Phis], P, [Phi|Result]) :- 
	pl_dp_one_literal_rule_remove_clauses(Phis, P, Result).

%%%
%%% pl_dp_one_literal_rule(+Phi, -Psi)
%%%
%%% Succeeds if Phi and Psi are equsatisfiable and Phi contains a clause with a single disjunct (literal) and
%%% Psi is Phi where this clause is removed along with the removal of the negation of the literal from all 
%%% other clauses.
%%%
%%% This is also called unit propagation.
%%%
pl_dp_one_literal_rule(Phi, Psi) :-
	member([P], Phi),
	pl_negate(P, NegP),
	pl_dp_one_literal_rule_remove_clauses(Phi, P, Mid),
	pl_dp_one_literal_rule_remove_negated_literals(Mid, NegP,Psi).

%%%
%%% pl_affirmitive_negative_rule_remove_clauses(+ClausesIn, +Pures, -ClausesOut)
%%%
%%% Succeeds if ClausesOut is ClausesIn with all clauses which contain an occurrence of a positive or negative 
%%% literal in Pures removed. Pures only contains positive literals.
%%% 
pl_affirmitive_negative_rule_remove_clauses([], _, []).
pl_affirmitive_negative_rule_remove_clauses([Phi|Phis], Pures, Result) :- 
	member(P, Pures),
	(member(P, Phi) ; member(~P, Phi)),
	!,
	pl_affirmitive_negative_rule_remove_clauses(Phis, Pures, Result).
pl_affirmitive_negative_rule_remove_clauses([Phi|Phis], Pures, [Phi|Result]) :- 
	pl_affirmitive_negative_rule_remove_clauses(Phis, Pures, Result).

%%%
%%% pl_dp_affirmative_negative_rule(Phi, Psi)
%%%
%%% Succeeds if Psi is Phi with all claues that contain a "pure" literal removed.
%%%
%%% We can delete any clause that contains a literal than only occurs either positively or
%%% negatively and not both. Such a literal is called "pure". This preserves satsfiablity.
%%%
%%% Note that for large input clauses you may want to replace the calls to setof with 
%%% somthing that doesn't copy between the clause store and the heap.
%%%
pl_dp_affirmative_negative_rule(Phi, Psi) :-
	setof(Pos, Clause^(member(Clause, Phi), member(Pos, Clause), Pos \= ~_), Positives),
	setof(Neg, Clause^(member(Clause, Phi), member(~Neg, Clause)), Negatives),
	ord_subtract(Positives, Negatives, OnlyPositives),
	ord_subtract(Negatives, Positives, OnlyNegatives),
	append(OnlyPositives, OnlyNegatives, Pures),
	Pures \= [],                                                  % must have them!
	pl_affirmitive_negative_rule_remove_clauses(Phi, Pures, Psi).

%%%
%%% pl_dpll_split_rule(ClausesIn, ClausesOut0, ClausesOut1)
%%%
%%% Succeeds if there is a literal L such that ClausesOut0 is ClausesIn union [L]
%%% and ClausesOut1 is ClausesIn union [~L].
%%%
%%% The 1-literal rule will then eliminate the unit clause and any clauses that 
%%% contain the literal L and remove any negations of L in clauses.
%%%
%%% This is the place to add in clever heuristics by choosing the most advantageous
%%% literal to plit on.
%%%
pl_dpll_split_rule(ClausesIn, [[Lit]|ClausesIn], [[LitNeg]|ClausesIn]) :-
	member(Clause, ClausesIn),
	member(Lit, Clause),
	pl_negate(Lit, LitNeg).

%%%
%%% pl_dpll_loop(+ClausesIn, -ClausesOut)
%%%
%%% Try to apply the following rules in this order to ClausesIn to give
%%% ClausesOut:
%%%   - 1-literal rule (unit propagation).
%%%   - Affirmative Negative rule.
%%%   - Split rule.
%%%
pl_dpll_loop(ClausesIn, ClausesOut) :-
	pl_dp_one_literal_rule(ClausesIn, ClausesMid),	
	!,
	pl_dpll_loop(ClausesMid, ClausesOut).
pl_dpll_loop(ClausesIn, ClausesOut) :-
	pl_dp_affirmative_negative_rule(ClausesIn, ClausesMid),
	!,
	pl_dpll_loop(ClausesMid, ClausesOut).
pl_dpll_loop(ClausesIn, ClausesOut) :-
	pl_dpll_split_rule(ClausesIn, ClausesMid0, ClausesMid1),
	!,
	(pl_dpll_loop(ClausesMid0, ClausesOut)
    ;
	pl_dpll_loop(ClausesMid1, ClausesOut)
        ).
pl_dpll_loop(Clauses, Clauses).

%%%
%%% pl_dpll_sat(+Phi)
%%%
%%% Succeeds if Phi is satisfiable.
%%%
pl_dpll_sat(Phi) :-
	cnf_transform(Phi, CNF),
	pl_clausal_form(CNF, Clauses),
	pl_dpll_loop(Clauses, []).

%%%
%%% pl_dpll_tautology(+Phi)
%%%
%%% Succeeds if Phi is a tautology.
%%%
pl_dpll_tautology(Phi) :-
	\+ pl_dpll_sat(~Phi).

