%-*-mode:prolog-*-

%%%
%%% pl_nenf.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the transformation of formulas in propositional logic
%%% to Negation Equivalence Normal Form (NENF).
%%%
:- ensure_loaded(examples(pl_operators)).

%%%
%%% nenf_rewrite_connectives(+A, ?B)
%%%
%%% The formula A is rewritten to B so that an implication
%%%   Phi => Psi
%%% is replaced by
%%%   ~Phi \/ Psi
%%%
nenf_rewrite_connectives(~A0, ~A1) :- 
	!,
	nenf_rewrite_connectives(A0, A1).
nenf_rewrite_connectives(A0 /\ B0, A1 /\ B1) :-
	!,
	nenf_rewrite_connectives(A0, A1),
	nenf_rewrite_connectives(B0, B1).
nenf_rewrite_connectives(A0 \/ B0, A1 \/ B1) :-
	!,
	nenf_rewrite_connectives(A0, A1),
	nenf_rewrite_connectives(B0, B1).
nenf_rewrite_connectives(A0 => B0, ~A1 \/ B1) :-
	!,
	nenf_rewrite_connectives(A0, A1),
	nenf_rewrite_connectives(B0, B1).
nenf_rewrite_connectives(A0 <=> B0, A1 <=> B1) :-
	!,
	nenf_rewrite_connectives(A0, A1),
	nenf_rewrite_connectives(B0, B1).
nenf_rewrite_connectives(A, A).

%%%
%%% nenf_push_negation(+A, -B)
%%%
%%% The formula A is rewritten to B using the following equivalences
%%%   ~(~A) <=> A
%%%   ~(A /\ B) <=> ~A \/ ~B  (De Morgan)
%%%   ~(A \/ B) <=> ~A /\ ~B  (De Morgan)
%%%   ~(A <=> B) <=> ~A <=> B 
%%%   ~false <=> true
%%%   ~true <=> false
%%%
nenf_push_negation(~(~A0), A1) :- 
	!,
	nenf_push_negation(A0, A1).
nenf_push_negation(~(A0 /\ B0), A1 \/ B1) :-
	!,
	nenf_push_negation(~A0, A1),
	nenf_push_negation(~B0, B1).
nenf_push_negation(~(A0 \/ B0), A1 /\ B1) :-
	!,
	nenf_push_negation(~A0, A1),
	nenf_push_negation(~B0, B1).
nenf_push_negation(~(A0 <=> B0), A1 <=> B1) :-
	!,
	nenf_push_negation(A0, A1),
	nenf_push_negation(~B0, B1).
nenf_push_negation(A0 /\ B0, A1 /\ B1) :-
	!,
	nenf_push_negation(A0, A1),
	nenf_push_negation(B0, B1).
nenf_push_negation(A0 \/ B0, A1 \/ B1) :-
	!,
	nenf_push_negation(A0, A1),
	nenf_push_negation(B0, B1).
nenf_push_negation(A0 <=> B0, A1 <=> B1) :-
	!,
	nenf_push_negation(A0, A1),
	nenf_push_negation(B0, B1).
nenf_push_negation(~ true, false).
nenf_push_negation(~ false, true).
nenf_push_negation(A, A).


%%%
%%% nenf_transform(+A, ?B)
%%%
%%% Formula B is a negation normal form of formula A.
%%%
nenf_transform(A, C) :-
	nenf_rewrite_connectives(A, B),
	nenf_push_negation(B, C).
	


