%-*-mode:prolog-*-

%%%
%%% pl_nnf.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the transformation of formulas in propositional logic
%%% to Negation Normal Form (NNF).
%%%
:- ensure_loaded(examples(pl_operators)).

%%%
%%% nnf_rewrite_connectives(+A, ?B)
%%%
%%% The formula A is rewritten to B so that an implication
%%%   Phi => Psi
%%% is replaced by
%%%   ~Phi \/ Psi
%%% and an equivalance
%%%   Phi <=> Psi
%%% is replaced by
%%%   (~Phi \/ Psi) /\ (Phi \/ ~Psi)
%%%
nnf_rewrite_connectives(~A0, ~A1) :- 
	!,
	nnf_rewrite_connectives(A0, A1).
nnf_rewrite_connectives(A0 /\ B0, A1 /\ B1) :-
	!,
	nnf_rewrite_connectives(A0, A1),
	nnf_rewrite_connectives(B0, B1).
nnf_rewrite_connectives(A0 \/ B0, A1 \/ B1) :-
	!,
	nnf_rewrite_connectives(A0, A1),
	nnf_rewrite_connectives(B0, B1).
nnf_rewrite_connectives(A0 => B0, ~A1 \/ B1) :-
	!,
	nnf_rewrite_connectives(A0, A1),
	nnf_rewrite_connectives(B0, B1).
nnf_rewrite_connectives(A0 <=> B0, (~A1 \/ B1) /\ (A1 \/ ~B1)) :-
	!,
	nnf_rewrite_connectives(A0, A1),
	nnf_rewrite_connectives(B0, B1).
nnf_rewrite_connectives(A, A).

%%%
%%% nnf_push_negation(+A, -B)
%%%
%%% The formula A is rewritten to B using the following equivalences
%%%   ~(~A) <=> A
%%%   ~(A /\ B) <=> ~A \/ ~B  (De Morgan)
%%%   ~(A \/ B) <=> ~A /\ ~B  (De Morgan)
%%%   ~false <=> true
%%%   ~true <=> false
%%%
nnf_push_negation(~(~A0), A1) :- 
	!,
	nnf_push_negation(A0, A1).
nnf_push_negation(~(A0 /\ B0), A1 \/ B1) :-
	!,
	nnf_push_negation(~A0, A1),
	nnf_push_negation(~B0, B1).
nnf_push_negation(~(A0 \/ B0), A1 /\ B1) :-
	!,
	nnf_push_negation(~A0, A1),
	nnf_push_negation(~B0, B1).
nnf_push_negation(A0 /\ B0, A1 /\ B1) :-
	!,
	nnf_push_negation(A0, A1),
	nnf_push_negation(B0, B1).
nnf_push_negation(A0 \/ B0, A1 \/ B1) :-
	!,
	nnf_push_negation(A0, A1),
	nnf_push_negation(B0, B1).
nnf_push_negation(~ true, false).
nnf_push_negation(~ false, true).
nnf_push_negation(A, A).


%%%
%%% nnf_transform(+A, ?B)
%%%
%%% Formula B is a negation normal form of formula A.
%%%
nnf_transform(A, C) :-
	nnf_rewrite_connectives(A, B),
	nnf_push_negation(B, C).
	


