%-*-mode:prolog-*-

%%%
%%% pl_tableau.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the tableau proof procedure for propositional
%%% logic.
%%%
:- ensure_loaded(examples(pl_operators)).

%%%
%%% pl_negate(+A, ?B)
%%%
%%% Succeeds if A is the negation of B.
%%%
pl_negate(true, false)  :- !.
pl_negate(false, true)  :- !.
pl_negate(~A, A) :- !.
pl_negate(A, ~A).

%%%
%%% pl_tableau_transform(+TreeIn, -TreeOut)
%%%
%%% Transform TreeIn into TreeOut. The representation of a tree is a
%%% list of complete branches. Each branch is a conjunction of its node's
%%% formulas.
%%%
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~ true, Branch, Rest),
	NewBranches = [[false|Rest]|Branches].
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~ false, Branch, Rest),
	NewBranches = [[true|Rest]|Branches].
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~(~A), Branch, Rest),
	NewBranches = [[A|Rest]|Branches].
pl_tableau_transform([Branch|Branches], [NewBranch|Branches]) :-
	del(A /\ B, Branch, Rest),
	NewBranch = [A, B|Rest].
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(A \/ B, Branch, Rest),
	NewBranches = [[A|Rest], [B|Rest]|Branches].
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(A => B, Branch, Rest),
	pl_tableau_transform([[~A \/ B|Rest]|Branches], NewBranches).
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(A <=> B, Branch, Rest),
	pl_tableau_transform([[(A => B) /\ (B => A)|Rest]|Branches], NewBranches).
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~(A /\ B), Branch, Rest),
	pl_tableau_transform([[~A \/ ~B|Rest]|Branches], NewBranches).
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~(A \/ B), Branch, Rest),
	pl_tableau_transform([[~A /\ ~B|Rest]|Branches], NewBranches).
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~(A => B), Branch, Rest),
	pl_tableau_transform([[A /\ ~B|Rest]|Branches], NewBranches).
pl_tableau_transform([Branch|Branches], NewBranches) :-
	del(~(A <=> B), Branch, Rest),
	pl_tableau_transform([[A <=> ~B|Rest]|Branches], NewBranches).

%%%
%%% pl_close_branches(+TreeIn, -TreeOut)
%%%
%%% TreeOut is TreeIn with all closed branches removed. A closed branch is a 
%%% contradiction. In our case that means either false, or, F /\ ~F for any 
%%% formula F.
%%%
pl_close_branches([], []).
pl_close_branches([H|T], Result) :-
	member(false, H),
	!,
	pl_close_branches(T, Result).
pl_close_branches([H|T], Result) :-
	member(A, H),
	pl_negate(A, NegA),
	member(NegA, H),
	!,
	pl_close_branches(T, Result).
pl_close_branches([H|T], [H|Result]) :-
	pl_close_branches(T, Result).

%%%
%%% pl_tableau_loop(+Tree)
%%%
%%% Succeeds if Tree can have all of its branches closed.
%%%
pl_tableau_loop([]) :- !.
pl_tableau_loop(Tree0) :- 
	pl_tableau_transform(Tree0, Tree1),
	!,
	pl_close_branches(Tree1, Tree2),
	pl_tableau_loop(Tree2).
pl_tableau_loop(Tree1) :- 
	pl_close_branches(Tree1, Tree2),
	pl_tableau_loop(Tree2).

%%%
%%% pl_tableau(+Formula)
%%%
%%% Succeeds if Formula can be refuted.
%%%
pl_tableau(Formula) :-
	pl_tableau_loop([[Formula]]).


%%%
%%% pl_tableau_tautology(+Formula)
%%%
%%% Succeeds if Formula is a tautology.
%%%
pl_tableau_tautology(Formula) :-
	pl_tableau(~Formula).

