%-*-mode:prolog-*-

%%%
%%% pl_truth_table.pl
%%%
%%% Copyright (C) 2014 Barry Watson. See LICENSE file for details.
%%%
%%% This file contains code for the generation of truth tables for 
%%% propositional logic.
%%%
:- ensure_loaded(examples(pl_operators)).
:- ensure_loaded(runtime(assoc)).
:- ensure_loaded(runtime(ordset)).

%%%
%%% pl_atoms(F, A).
%%%
%%% A is the set of atoms in the formula F.
%%%
pl_atoms(Atom, [Atom]) :- 
	atom(Atom), 
	!.
pl_atoms(~Phi, Atoms) :- 
	pl_atoms(Phi, Atoms).
pl_atoms(Phi /\ Psi, Atoms) :- 
	pl_atoms(Phi, A1), 
	pl_atoms(Psi, A2), 
	ord_union(A1, A2, Atoms).
pl_atoms(Phi \/ Psi, Atoms) :- 
	pl_atoms(Phi, A1), 
	pl_atoms(Psi, A2), 
	ord_union(A1, A2, Atoms).
pl_atoms(Phi => Psi, Atoms) :- 
	pl_atoms(Phi, A1), 
	pl_atoms(Psi, A2), 
	ord_union(A1, A2, Atoms).
pl_atoms(Phi <=> Psi, Atoms) :- 
	pl_atoms(Phi, A1), 
	pl_atoms(Psi, A2), 
	ord_union(A1, A2, Atoms).

%%%
%%% pl_eval(F, V, Answer)
%%%
%%% Using the valuation (truth assignment) V we evaluate the formula F.
%%% This is the mathematical definition of the semantics of propositional logic.
%%% An association list is used to implement the truth assignment.
%%%
%%% assoc_to_list(V, [p-true, q-false, r-true]), pl_eval((p /\ q => q \/ r), V).
%%% assoc_to_list(V, [p-true, q-true, r-false]), pl_eval((p /\ q => q /\ r), V).
%%%
pl_eval(Atom, V) :- 
	atom(Atom),  
	(get_assoc(Atom, V, Answer); throw(no_such_proposition_variable(Atom))),
	!,
	Answer = true.
pl_eval(~Phi, V) :- 
	\+ pl_eval(Phi, V).
pl_eval(Phi /\ Psi, V) :- 
	pl_eval(Phi, V), 
	pl_eval(Psi, V).
pl_eval(Phi \/ Psi, V) :- 
	pl_eval(Phi, V)
    ; 
	pl_eval(Psi, V).
pl_eval(Phi => Psi, V) :- 
	pl_eval(~Phi \/ Psi, V).
pl_eval(Phi <=> Psi, V) :-
	pl_eval((~Phi \/ Psi) /\ (Phi \/ ~Psi), V).


%%%
%%% pl_valuation(AtomList, V)
%%%
%%% V is a valuation of the atoms in AtomList.
%%%
pl_valuation([], t).                      % The empty association is denoted by the symbol t. 
pl_valuation([H|T], V) :-
	member(Truth, [false, true]),     % Do this first so we create a choice point for the first element before the last element.
	pl_valuation(T, Vtemp),
	put_assoc(H, Vtemp, Truth, V).

%%%
%%% pl_print_truth_table_row(Cols, T)
%%%
%%% Print one row of a truth table. Cols is a set of true or false,
%%% one for each variable of the table. T is either true or false
%%% depending on the evaluation of the original formula w.r.t. the 
%%% valuation represented by the values in Cols.
%%%
pl_print_truth_table_row([], Truth) :- 
	print(Truth), 
	nl.
pl_print_truth_table_row([(_-X)|T], Truth) :- 
	print(X), 
	put_code(9),   % ASCII 9 is a horizontal TAB.
	pl_print_truth_table_row(T, Truth).

%%%
%%% pl_print_truth_table_head(Cols)
%%%
%%% Print the head row of the truth table. There
%%% is one element in Cols for each variable.
%%%
pl_print_truth_table_head([]) :- 
	print(formula),
	nl.
pl_print_truth_table_head([H|T]) :- 
	print(H), 
	put_code(9),   % ASCII 9 is a horizontal TAB.
	pl_print_truth_table_head(T).

%%%
%%% pl_print_truth_table(F)
%%%
%%% Print a truth table for the formula F.
%%%
%%% pl_print_truth_table(((p => q) => p) => p).
%%% pl_print_truth_table((p /\ q => q /\ r)).
%%%
pl_print_truth_table(Phi) :-
	print('Truth table for '), 
	print(Phi),
	nl,
	pl_atoms(Phi, Atoms),
	pl_print_truth_table_head(Atoms),
	pl_print_truth_table(Atoms, Phi).

pl_print_truth_table(Atoms, Phi) :-
	pl_valuation(Atoms, V),                % The failure driven loop starts here. Create a valuation V
	(   pl_eval(Phi, V) ->                 % Evaluate the formula given V
	    Truth = true
	;
	    Truth = false
	),
	assoc_to_list(V, Lst),                 % Turn the valuation into a list of true/false elements (a row).
	pl_print_truth_table_row(Lst, Truth),  % Print the row.
	fail.                                  % Force a new round of the failure driven loop, i.e., create a new valuation.
pl_print_truth_table(_, _).                    % Finished.

